#include "StdAfx.h"
#include "FLVariable.h"

int CFLVariable::VarCounter = 0;	// static var

CFLVariable::CFLVariable(int NumStates, BOOL bIsInput)
{
	++VarCounter;				// Variables are given a unique number
	m_nNumStates = NumStates;
	m_bIsInput = bIsInput;
	m_cFuzzySet = new CFuzzySet(NumStates);	// Caution: Fuzzy Set is not defined in this constructor
}

CFLVariable::CFLVariable(CFuzzySet &fs, BOOL bIsInput)	// copy const.
{
	++VarCounter;
	m_nNumStates = fs.m_nNumMFs;
	m_bIsInput = bIsInput;
	m_cFuzzySet = new CFuzzySet(m_nNumStates);
	// member for member copy (could be a temp solution)
	for (int i=0;i<m_nNumStates;i++)
	{
		++(m_cFuzzySet->FuzzyState[i].m_nPremiseCounter);
		m_cFuzzySet->FuzzyState[i].m_nPremiseNumber = m_cFuzzySet->FuzzyState[i].m_nPremiseCounter;
        m_cFuzzySet->FuzzyState[i].DOM = 0.0;
        m_cFuzzySet->FuzzyState[i].m_nType  = fs.FuzzyState[i].m_nType;
        m_cFuzzySet->FuzzyState[i].m_dPoint1 = fs.FuzzyState[i].m_dPoint1;
        m_cFuzzySet->FuzzyState[i].m_dPoint2 = fs.FuzzyState[i].m_dPoint2;
		m_cFuzzySet->FuzzyState[i].m_dCenter = fs.FuzzyState[i].m_dCenter;
		m_cFuzzySet->FuzzyState[i].m_strName = fs.FuzzyState[i].m_strName;
	}
}

CFLVariable::CFLVariable(void)		// default constructor
{									// no fuzzy set here attached
	++VarCounter;
	m_cFuzzySet = NULL;
}

CFLVariable::~CFLVariable(void)
{
	UnloadFuzzySet();
}

int CFLVariable::LoadFuzzySet(CFuzzySet &fs, BOOL bIsInput)
{
	if (m_cFuzzySet != NULL) return -1;	// do not load over existing Fuzzy Set
	m_nNumStates = fs.m_nNumMFs;
	m_bIsInput = bIsInput;
	m_cFuzzySet = new CFuzzySet(m_nNumStates);
	// member for member copy (could be a temp solution)
	for (int i=0;i<m_nNumStates;i++)
	{
		++(m_cFuzzySet->FuzzyState[i].m_nPremiseCounter);
		m_cFuzzySet->FuzzyState[i].m_nPremiseNumber = m_cFuzzySet->FuzzyState[i].m_nPremiseCounter;
        m_cFuzzySet->FuzzyState[i].DOM = 0.0;
        m_cFuzzySet->FuzzyState[i].m_nType  = fs.FuzzyState[i].m_nType;
        m_cFuzzySet->FuzzyState[i].m_dPoint1 = fs.FuzzyState[i].m_dPoint1;
        m_cFuzzySet->FuzzyState[i].m_dPoint2 = fs.FuzzyState[i].m_dPoint2;
		m_cFuzzySet->FuzzyState[i].m_dCenter = fs.FuzzyState[i].m_dCenter;
		m_cFuzzySet->FuzzyState[i].m_strName = fs.FuzzyState[i].m_strName;
	}
	return 0;
}

void CFLVariable::UnloadFuzzySet(void)
{
	if (m_cFuzzySet != NULL) delete m_cFuzzySet;
	m_cFuzzySet = NULL;
}