// FuzzyDlg.cpp : implementation file
//

//	This is an implementation of a FL Controller as per the 
//	SeattleRobotics.org tutorial
//	http://www.seattlerobotics.org/encoder/mar98/fuz/flindex.htm
//
//	Free to distribute and copy.  
//	Mention of the author and original tutorial must remain visible in client application
//
//	BV 2003


#include "stdafx.h"
#include "FuzzyDlg.h"


#include "math.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#endif

#define NOTHING -1
#define AND 0
#define OR 1
#define NOT 2

// Forward declarations
	//double FuzzyController(double,double);
	void LoadFuzzySets(CFuzzySet &fs1, CFuzzySet &fs2, CFuzzySet &fs3);
	void LoadFLRuleSet(CFLRule *rs);
	void CalcPremises(CFLVariable &Var, double Value);
	double And(double x, double y);
	double Or(double x, double y);
	double Not(double x);

// CAboutDlg dialog used for App About

class CAboutDlg : public CDialog
{
public:
	CAboutDlg();

// Dialog Data
	enum { IDD = IDD_ABOUTBOX };

	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support

// Implementation
protected:
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
END_MESSAGE_MAP()


// CFuzzyDlg dialog



CFuzzyDlg::CFuzzyDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CFuzzyDlg::IDD, pParent)
	, Input1(0)
	, Input2(0)
	, Output(0)
{

	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
}

CFuzzyDlg::~CFuzzyDlg()
{
}

void CFuzzyDlg::DoDataExchange(CDataExchange* pDX)
{
	DDX_Text(pDX,IDC_EDIT1,Input1);
	DDX_Text(pDX,IDC_EDIT2,Input2);
	CDialog::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_LIST1, lstPredicate);
	DDX_Control(pDX, IDC_LIST3, lstOutputStrength);
	DDX_Control(pDX, IDC_LIST4, lstEvalPredicates);
	DDX_Control(pDX, IDC_EDIT3, m_edtOutput);
}



BEGIN_MESSAGE_MAP(CFuzzyDlg, CDialog)
	ON_WM_SYSCOMMAND()
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	//}}AFX_MSG_MAP
	ON_BN_CLICKED(IDOK, OnBnClickedOk)
	ON_BN_CLICKED(IDCANCEL, OnBnClickedCancel)
END_MESSAGE_MAP()


// CFuzzyDlg message handlers

BOOL CFuzzyDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	SetWindowText("Simple Fuzzy Controller (BV 2003)");

	// Add "About..." menu item to system menu.

	// IDM_ABOUTBOX must be in the system command range.
	ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
	ASSERT(IDM_ABOUTBOX < 0xF000);

	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pSysMenu != NULL)
	{
		CString strAboutMenu;
		strAboutMenu.LoadString(IDS_ABOUTBOX);
		if (!strAboutMenu.IsEmpty())
		{
			pSysMenu->AppendMenu(MF_SEPARATOR);
			pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, strAboutMenu);
		}
	}

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

	Input1 = 0.0;
	Input2 = 0.0;
	return TRUE;  // return TRUE  unless you set the focus to a control
}


void CFuzzyDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
	if ((nID & 0xFFF0) == IDM_ABOUTBOX)
	{
		CAboutDlg dlgAbout;
		dlgAbout.DoModal();
	}
	else
	{
		CDialog::OnSysCommand(nID, lParam);
	}
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CFuzzyDlg::OnPaint() 
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialog::OnPaint();
	}
}

// The system calls this function to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CFuzzyDlg::OnQueryDragIcon()
{
	return static_cast<HCURSOR>(m_hIcon);
}

void CFuzzyDlg::OnBnClickedOk()
{
	CString str;
	lstPredicate.ResetContent();
	lstEvalPredicates.ResetContent();
	lstOutputStrength.ResetContent();
	UpdateData(TRUE);
	Output = FuzzyController(Input1,Input2);
	str.Format("%.3f",Output);
	m_edtOutput.SetWindowText(str);
}

void CFuzzyDlg::OnBnClickedCancel()
{
	OnCancel();
}

//  Functions

//	This is a sample function, in absence of a graphical UI.
//	There are 3 Fuzzy Sets on this example

//	This function must be reviewed depending on UI yet to be coded
void LoadFuzzySets(CFuzzySet &fs1, CFuzzySet &fs2, CFuzzySet &fs3)
{	
	// Point1 and Point2 are used to build the MF shape
	// One can also add bell shape type, with Point1 being average, and
	// Point2 being width

	// Center is generally the actual center of the region
	// For shoulder type shapes, one uses the range boundary (-1 or +1)

	// point value i.e. x-coord are limited to [-1,1]. Scaling is assumed.
	// fs1 ShoulderRight	// -\	"Neg"  Center set to -1 (min)
	fs1.FuzzyState[0].m_nType = 0;			
	fs1.FuzzyState[0].m_dPoint1 = -1.0/2.0; 
	fs1.FuzzyState[0].m_dPoint2 = 0.0;		
	fs1.FuzzyState[0].m_dCenter = -1.0;	
	fs1.FuzzyState[0].m_strName = "FS1 - Neg";

	// fs1 Triangle			// /\	"Null" Center set to triangle center = 0
	fs1.FuzzyState[1].m_nType = 1;
	fs1.FuzzyState[1].m_dPoint1 = -1.0/2.0;
	fs1.FuzzyState[1].m_dPoint2 = 1.0/2.0;
	fs1.FuzzyState[1].m_dCenter = 0.0;
	fs1.FuzzyState[1].m_strName = "FS1 - Null";
	//fs1.FuzzyState[1].m_dCenter = (fs1.FuzzyState[1].m_dPoint1 + fs1.FuzzyState[1].m_dPoint2)/2.0;

	// fs1 ShoulderLeft		// /-	"Positive" Center set to +1 (max)
	fs1.FuzzyState[2].m_nType = 2;
	fs1.FuzzyState[2].m_dPoint1 = 0.0;
	fs1.FuzzyState[2].m_dPoint2 = 1.0/2.0;
	fs1.FuzzyState[2].m_dCenter = 1.0;		
	fs1.FuzzyState[2].m_strName = "FS1 - Pos";

	// fs2 ShoulderRight
	fs2.FuzzyState[0].m_nType = 0;
	fs2.FuzzyState[0].m_dPoint1 = -1.0/2.0;
	fs2.FuzzyState[0].m_dPoint2 = 0.0;
	fs2.FuzzyState[0].m_dCenter = -1.0;
	fs2.FuzzyState[0].m_strName = "FS2 - Neg";

	// fs2 Triangle
	fs2.FuzzyState[1].m_nType = 1;
	fs2.FuzzyState[1].m_dPoint1 = -1.0/2.0;
	fs2.FuzzyState[1].m_dPoint2 = 1.0/2.0;
	fs2.FuzzyState[1].m_dCenter = 0.0;
	fs2.FuzzyState[1].m_strName = "FS2 - Null";
	//fs2.FuzzyState[1].m_dCenter = (fs1.FuzzyState[1].m_dPoint1 + fs1.FuzzyState[1].m_dPoint2)/2.0;

	// fs2 ShoulderLeft
	fs2.FuzzyState[2].m_nType = 2;
	fs2.FuzzyState[2].m_dPoint1 = 0.0;
	fs2.FuzzyState[2].m_dPoint2 = 1.0/2.0;
	fs2.FuzzyState[2].m_dCenter = 1.0;		
	fs2.FuzzyState[2].m_strName = "FS2 - Pos";

	// fs3 ShoulderRight
	fs3.FuzzyState[0].m_nType = 0;
	fs3.FuzzyState[0].m_dPoint1 = -1.0/2.0;
	fs3.FuzzyState[0].m_dPoint2 = 0.0;
	fs3.FuzzyState[0].m_dCenter = -1.0;	
	fs3.FuzzyState[0].m_strName = "FS3 - Neg";

	// fs3 Triangle 
	fs3.FuzzyState[1].m_nType = 1;
	fs3.FuzzyState[1].m_dPoint1 = -1.0/2.0;
	fs3.FuzzyState[1].m_dPoint2 = 1.0/2.0;
	fs3.FuzzyState[1].m_dCenter = 0.0;
	fs3.FuzzyState[1].m_strName = "FS3 - Null";
	//fs3.FuzzyState[1].m_dCenter = (fs1.FuzzyState[1].m_dPoint1 + fs1.FuzzyState[1].m_dPoint2)/2.0;

	// fs3 ShoulderLeft
	fs3.FuzzyState[2].m_nType = 2;
	fs3.FuzzyState[2].m_dPoint1 = 0.0;
	fs3.FuzzyState[2].m_dPoint2 = 1.0/2.0;
	fs3.FuzzyState[2].m_dCenter = 1.0;
	fs3.FuzzyState[2].m_strName = "FS3 - Pos";
}


//	Rule predicates are built with premise numbers
//  as recorded while associating a variable to a a fuzzy set


// This is a very simple example where the top FL object (FLC) is
// not part of the dialog class

void LoadFLRuleSet(CFLRule *rs)
{	// No rule editor... Number of premises per predicate is here hard-coded.
	// Since fuzzy states are already defined with the fuzzy sets
	// one just needs to associate rule premises with existing states
	for (int i=0;i<rs->NumRules;i++)
	{
		switch (rs[i].m_nRuleNumber )
		{
			case 0:
				rs[i].m_pnPremise[0] = 1;			// First variable is "Negative"
				rs[i].m_pnOperation[0] = AND;
				rs[i].m_pnPremise[1] = 4;			// 2nd variable is "Negative"
				rs[i].m_pnOperation[1] = NOTHING;
				rs[i].m_nConsequent = 7;			// Output is "Negative"
				break;
			case 1:
				rs[i].m_pnPremise[0] = 2;			// First variable is "Null"
				rs[i].m_pnOperation[0] = AND;
				rs[i].m_pnPremise[1] = 4;			// 2nd variable is "Negative"
				rs[i].m_pnOperation[1] = NOTHING;
				rs[i].m_nConsequent = 9;			// Output is "Positive"
				break;
			case 2:
				rs[i].m_pnPremise[0] = 3;			// First variable is "Positive"
				rs[i].m_pnOperation[0] = AND;
				rs[i].m_pnPremise[1] = 4;			// 2nd variable is "Negative"
				rs[i].m_pnOperation[1] = NOTHING;
				rs[i].m_nConsequent = 9;			// Output is "Positive"
				break;
			case 3:
				rs[i].m_pnPremise[0] = 1;			// First variable is "Negative"
				rs[i].m_pnOperation[0] = AND;
				rs[i].m_pnPremise[1] = 5;			// 2nd variable is "Null"
				rs[i].m_pnOperation[1] = NOTHING;
				rs[i].m_nConsequent = 7;			// Output is "Negative"
				break;
			case 4:
				rs[i].m_pnPremise[0] = 2;			// First variable is "Null"
				rs[i].m_pnOperation[0] = AND;
				rs[i].m_pnPremise[1] = 5;			// 2nd variable is "Null"
				rs[i].m_pnOperation[1] = NOTHING;
				rs[i].m_nConsequent = 8;			// Output is "Null"
				break;
			case 5:
				rs[i].m_pnPremise[0] = 3;			// First variable is "Positive"
				rs[i].m_pnOperation[0] = AND;
				rs[i].m_pnPremise[1] = 5;			// 2nd variable is "Null"
				rs[i].m_pnOperation[1] = NOTHING;
				rs[i].m_nConsequent = 9;			// Output is "Positive"
				break;
			case 6:
				rs[i].m_pnPremise[0] = 1;			// First variable is "Negative"
				rs[i].m_pnOperation[0] = AND;
				rs[i].m_pnPremise[1] = 6;			// 2nd variable is "Positive"
				rs[i].m_pnOperation[1] = NOTHING;
				rs[i].m_nConsequent = 7;			// Output is "Negative"
				break;
			case 7:
				rs[i].m_pnPremise[0] = 2;			// First variable is "Null"
				rs[i].m_pnOperation[0] = AND;
				rs[i].m_pnPremise[1] = 6;			// 2nd variable is "Positive"
				rs[i].m_pnOperation[1] = NOTHING;
				rs[i].m_nConsequent = 7;			// Output is "Negative"
				break;
			case 8:
				rs[i].m_pnPremise[0] = 3;			// First variable is "Positive"
				rs[i].m_pnOperation[0] = AND;
				rs[i].m_pnPremise[1] = 6;			// 2nd variable is "Positive"
				rs[i].m_pnOperation[1] = NOTHING;
				rs[i].m_nConsequent = 9;			// Output is "Positive"
				break;
			default:	// should not get here...
				rs[i].m_pnPremise[0] = 0;			
				rs[i].m_pnOperation[0] = NOTHING;
				rs[i].m_pnPremise[1] = 0;			
				rs[i].m_pnOperation[1] = NOTHING;
				rs[i].m_nConsequent = 0;
				break;
		}
	}
}

//	A reference to the rule set and a pointer to the variable array
//  is given
double CFuzzyDlg::EvaluateRulePredicate(CFLRule &rs, CFLVariable *FV)
{
	CString strDisplay;
	double value=0.0;
	// An array of doubles will contain each premise evaluation
	// Then the overall Predicate is calculated
	double *RuleEstimate = new double[rs.m_nNumPremises];
	for (int i=0;i<rs.m_nNumPremises;i++)
		// Premise numbers are unique
		// One just needs to find which variable it can be associated with
		for (int j=0;j<FV[0].VarCounter;j++)		// NumVariables = 3 in this example
		{
			if (FV[j].m_bIsInput)	// Only rule antecedents 
				for (int k=0;k<FV[j].m_nNumStates;k++)	// Num FuzzyStates for each variable = 3
					if (rs.m_pnPremise[i] == FV[j].m_cFuzzySet->FuzzyState[k].m_nPremiseNumber  )
						RuleEstimate[i] = FV[j].m_cFuzzySet->FuzzyState[k].DOM;
		}
	// Now apply "fuzzy" booleans to Premises to evaluate the Predicate
	for (int i=0;i<rs.m_nNumPremises;i++)
	{
		switch (rs.m_pnOperation[i])
		{
			case AND:
				value += And(RuleEstimate[i],RuleEstimate[i+1]);
				strDisplay.Format("AND(%.2f,%.2f)=%.2f",RuleEstimate[i],RuleEstimate[i+1],value);
				break;
			case OR:
				value += Or(RuleEstimate[i],RuleEstimate[i+1]);
				strDisplay.Format("OR(%.2f,%.2f)=%.2f",RuleEstimate[i],RuleEstimate[i+1],value);
				strDisplay = "OR(";
				break;
			case NOT:
				value += Not(RuleEstimate[i]);
				strDisplay.Format("NOT(%.2f)=%.2f",RuleEstimate[i],value);
				break;
			case NOTHING:
				// End predicate evaluation
			default:
				;
		}
	}

	lstEvalPredicates.AddString(strDisplay);

	delete [] RuleEstimate;
	RuleEstimate = NULL;
	return value;
}

//	Calculating Premises, i.e the DOM of a variable in each possible fuzzy state
void CalcPremises(CFLVariable &Var, double Value)
{
	// Var.m_cFuzzySet->m_nNumMFs = number of states
	for (int i=0;i<Var.m_cFuzzySet->m_nNumMFs;i++)
		Var.m_cFuzzySet->FuzzyState[i].CalcDOM(Value);
}

double Or(double x, double y)
{	if (x>y) return x; else return y;	}	// MAX

double And(double x, double y)
{	if (x>y) return y; else return x;	}	// MIN

double Not(double x)
{	return (1-x);	}


// Main FL routine

double CFuzzyDlg::FuzzyController(double Value1, double Value2)
{
	// FUZZY CONTROLLER
	
	//  Fuzzy Set setup : 3 fuzzy sets
	//  which are then associated to each variable (2 inputs, 1 output)
	//	The number of fuzzy states can also be set in the constructor.
	double OutputValue = 0.0;

	CFLC FLC;		// instantiate a new FL controller

	FLC.InitFuzzyStates(3);		// 3 Fuzzy Sets		
	// Since we used the default constructor, we need to initialise the Fuzzy Sets
	FLC.FS[0].InitFuzzySet(3);	// 3 states defined (Neg,Null,Pos)
	FLC.FS[1].InitFuzzySet(3);
	FLC.FS[2].InitFuzzySet(3);

	//	Exact settings of membership functions are here defined (or dumped here i should say...)
	//	Would be improved for scalability using a FS editor
	LoadFuzzySets(FLC.FS[0],FLC.FS[1],FLC.FS[2]);

	// 2 inputs, 1 output: each is associated with a fuzzy set
	// defined above
	FLC.InitVarSet(3);
	FLC.FLV[0].LoadFuzzySet(FLC.FS[0],TRUE);	// This is Input1 in the FL Controller
	FLC.FLV[1].LoadFuzzySet(FLC.FS[1],TRUE);	// This is Input2 
	FLC.FLV[2].LoadFuzzySet(FLC.FS[2],FALSE);	// This is the rule output

	// Since the 'variable <-> FS' is done, there is no need
	// in this simple FL controller to establish a separate Premise object.
	// They are already all there.

	// Input values Value1 and Value2 come from the dialog box
	// Calculate the premise values
	CalcPremises(FLC.FLV[0],Value1);
	CalcPremises(FLC.FLV[1],Value2);

	// Display predicate values
	CString strVal;
	for (int i=0;i<2;i++)
	{
		for (int j=0;j<3;j++)
		{
	        strVal.Format("V%i %s = %.2f",
			i,
			FLC.FLV[i].m_cFuzzySet->FuzzyState[j].m_strName,
			FLC.FLV[i].m_cFuzzySet->FuzzyState[j].DOM);
			lstPredicate.AddString(strVal);
		}
	}

	// Premises are numbered:
	// Each fuzzy set contains 3 states here
	// So FuzzyVariable[0]	: premises 1 to 3 (from negative to positive)
	// FuzzyVariable[1]		: premises 4 to 6
	// FuzzyVariable[2]		: premises 7 to 9 (Maybe not rigorous to call them premises for a Consequent)

	// Build rule set
	// Rules are composed of a Predicate block followed by a Consequent Block
	// i.e. If Predicate then Consequent
	// with Predicate = Premise1 AND/OR Premise2 ...

	int NumRules = 9;	// 3 x 3 rule matrix (eq. to Fuzzy Sets)
	FLC.InitRuleSet(NumRules);
	// In this example, rule predicates are all made of 2 premises
	// i.e. If a AND b THEN
	for (int i=0;i<NumRules;i++) FLC.FLRS[i].Allocate(2);
	
	// Now the rule set can be loaded in the FLC
	// Again this is just an example of a rule set as per the SeattleRobotics tutorial
	LoadFLRuleSet(FLC.FLRS);

	//	Now, since premises have already been evaluated previously
	//	all is left to do really is inferencing the rules using input values
	double *PredicateValue = new double[NumRules];
	int	*Consequent = new int[NumRules];

	// Display
	for (int i=0;i<NumRules;i++)
	{
		strVal.Format("Rule:%i",i+1);
		lstEvalPredicates.AddString(strVal);
		PredicateValue[i] = EvaluateRulePredicate(FLC.FLRS[i],FLC.FLV);
		//strVal.Format("Rule%i = %.2f",i+1,PredicateValue[i]);
		Consequent[i] = FLC.FLRS[i].m_nConsequent;	// will be used next for defuzz
	}

	// And now defuzzify into a crisp output: Apply RSS (Root-Sum-Square) method, or fuzzy centroid
	// There are several more or less selective methods to apply predicate values
	// RSS is probably the most balanced of them.

	// The Output variable is the last one, here no 2, i.e. FuzzyVariable[2]
	// Generally, it will be available through: FuzzyVariable[0].VarCounter - 1;
	int NumOutputStates = FLC.FLV[2].m_cFuzzySet->m_nNumMFs;
	
	// After inferencing:
	// OutputStrength stores the "strength" in each output fuzzy state.
	double *OutputStrength = new double[NumOutputStates];
	for (int i=0;i<NumOutputStates;i++) OutputStrength[i] = 0.0;
	
	// FLV[2] is the output in this example
	// Premises have been numbered incrementally
	int FirstOutputState = FLC.FLV[2].m_cFuzzySet->FuzzyState[0].m_nPremiseNumber;
	int LastOutputState = FLC.FLV[2].m_cFuzzySet->FuzzyState[NumOutputStates-1].m_nPremiseNumber;
	double denom_out = 0.0;
	for (int i=FirstOutputState;i<=LastOutputState;i++)
	{	// Scanning through rule set to evaluate OutputStrength for each output fuzzy state
		double sumsq = 0.0;
		for (int j=0;j<NumRules;j++)
			if (Consequent[j]==i)	// Sum squares
				sumsq += PredicateValue[j] * PredicateValue[j];
		int k = i-FirstOutputState;
		OutputStrength[k] = sqrt(sumsq);
		// Display
		strVal.Format("%i %s %.2f",
			k,
			FLC.FLV[2].m_cFuzzySet->FuzzyState[k].m_strName, 
			OutputStrength[k]);
		lstOutputStrength.AddString(strVal);
		denom_out += OutputStrength[k];	// root-sum-squares
	}

	// Calculate the Fuzzy Centroid of the area (cf tutorial for details)
	double num_out = 0.0;
	for (int i=0;i<NumOutputStates;i++)
	{
		double dCenter;
		dCenter = FLC.FLV[2].m_cFuzzySet->FuzzyState[i].m_dCenter;
		num_out += dCenter * OutputStrength[i];
	}

	if (denom_out == 0.0)		OutputValue = 0.0;
	else						OutputValue = num_out / denom_out;

	//  That's it!  Time to clean up now...
	// reset static variables
	FLC.nFLC_Counter = 0;
	FLC.FLV[0].VarCounter = 0;
	FLC.FLV[0].m_cFuzzySet->FuzzyState[0].m_nPremiseCounter = 0;
	FLC.FLRS[0].NumRules = 0;

	// deallocate from the heap
	if (OutputStrength != NULL)
	{
		delete [] OutputStrength;
		OutputStrength = NULL;
	}
	if (Consequent != NULL)
	{
		delete [] Consequent;
		Consequent = NULL;
	}
	if (PredicateValue != NULL)
	{
		delete [] PredicateValue;
		PredicateValue = NULL;
	}
	// The rest will be cleaned in the FLC destructor
	return OutputValue;	// returns 'fuzzy' output
}
